/* -*- Mode:C++; c-file-style:"gnu"; indent-tabs-mode:nil; -*- */
/*
 * Copyright (c) 2010 Telum (www.telum.ru)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Author: Kirill Andreev <andreev@telum.ru>
 */

#include "ns3/pointer.h"
#include "ns3/uinteger.h"
#include "ns3/log.h"
#include "lrr-device.h"

NS_LOG_COMPONENT_DEFINE ("LrrNeighborAwareDevice");

namespace ns3 {
namespace lrr {

NS_OBJECT_ENSURE_REGISTERED (NeighborAwareDevice);

TypeId
NeighborAwareDevice::GetTypeId ()
{
  static TypeId tid = TypeId ("ns3::lrr::NeighborAwareDevice")
    .SetParent<NetDevice> ()
    .AddConstructor<NeighborAwareDevice> ()
    .AddAttribute ("Mac", "The MAC layer attached to this device.",
                   PointerValue (),
                   MakePointerAccessor (&NeighborAwareDevice::GetMac,
                                        &NeighborAwareDevice::SetMac),
                   MakePointerChecker<Mac> ())
  ;
  return tid;
}

NeighborAwareDevice::NeighborAwareDevice () :
  m_mac (0), m_node (0), m_ifIndex (0), m_mtu (0)
{}

void
NeighborAwareDevice::DoStart ()
{
  m_mac->Start ();
  NetDevice::DoStart ();
}

NeighborAwareDevice::~NeighborAwareDevice ()
{
  NS_LOG_FUNCTION (this);
}

void
NeighborAwareDevice::DoDispose ()
{
  NS_LOG_FUNCTION (this);
  m_node = 0;
  m_mac = 0;
  m_forwardUp = MakeNullCallback <bool,Ptr<NetDevice>,Ptr<const Packet>,uint16_t,const Address &> ();
  m_promiscRx = MakeNullCallback<bool, Ptr<NetDevice>, Ptr<const Packet>, uint16_t, const Address &, const Address &, enum PacketType> ();
  // chain up.
  NetDevice::DoDispose ();
}
// MAC {
void
NeighborAwareDevice::SetMac (Ptr<Mac> mac)
{
  m_mac = mac;
  m_mac->SetForwardUpCallback (MakeCallback (&NeighborAwareDevice::ForwardUp, this));
}

Ptr<Mac>
NeighborAwareDevice::GetMac () const
{
  return m_mac;
}
// }
// Inherited from NetDevice base class {
void
NeighborAwareDevice::SetIfIndex (const uint32_t index)
{
  m_ifIndex = index;
}

uint32_t
NeighborAwareDevice::GetIfIndex () const
{
  return m_ifIndex;
}

Ptr<Channel>
NeighborAwareDevice::GetChannel () const
{
  return 0;
}

void
NeighborAwareDevice::SetAddress (Address address)
{
  m_mac->SetAddress (address);
}

Address
NeighborAwareDevice::GetAddress () const
{
  NS_ASSERT (m_mac != 0);
  return m_mac->GetAddress ();
}

bool
NeighborAwareDevice::SetMtu (const uint16_t mtu)
{
  UintegerValue maxMsduSize;
  m_mac->GetAttribute ("MaxMsduSize", maxMsduSize);
  if (mtu > maxMsduSize.Get () || mtu == 0)
    {
      return false;
    }
  m_mtu = mtu;
  return true;
}

uint16_t
NeighborAwareDevice::GetMtu () const
{
  if (m_mtu == 0)
    {
      UintegerValue maxMsduSize;
      m_mac->GetAttribute ("MaxMsduSize", maxMsduSize);
      m_mtu = maxMsduSize.Get ();
    }
  return m_mtu;
}

bool
NeighborAwareDevice::IsLinkUp () const
{
  return true;
}

void
NeighborAwareDevice::AddLinkChangeCallback (Callback<void> callback)
{}

bool
NeighborAwareDevice::IsBroadcast () const
{
  return true;
}

Address
NeighborAwareDevice::GetBroadcast () const
{
  return m_mac->GetBroadcast ();
}

bool
NeighborAwareDevice::IsMulticast () const
{
  return true;
}

Address
NeighborAwareDevice::GetMulticast (Ipv4Address multicastGroup) const
{
  return m_mac->GetMulticast (multicastGroup);
}

Address
NeighborAwareDevice::GetMulticast (Ipv6Address multicastGroup) const
{
  return m_mac->GetMulticast (multicastGroup);
}

bool
NeighborAwareDevice::IsPointToPoint () const
{
  return false;
}

bool
NeighborAwareDevice::IsBridge () const
{
  return false;
}

bool
NeighborAwareDevice::Send (Ptr<Packet> packet, const Address& dest, uint16_t protocolNumber)
{
  m_mac->Enqueue (packet, dest, m_mac->GetAddress (), protocolNumber);
  return true;
}

bool
NeighborAwareDevice::SendFrom (Ptr<Packet> packet, const Address& source, const Address& dest, uint16_t protocolNumber)
{
  NS_FATAL_ERROR ("Radio does not support SendFrom");
  return false;
}

Ptr<Node>
NeighborAwareDevice::GetNode () const
{
  return m_node;
}

void
NeighborAwareDevice::SetNode (Ptr<Node> node)
{
  m_node = node;
}

bool
NeighborAwareDevice::NeedsArp () const
{
  return true;
}

void
NeighborAwareDevice::SetReceiveCallback (NetDevice::ReceiveCallback cb)
{
  m_forwardUp = cb;
}

void
NeighborAwareDevice::SetPromiscReceiveCallback (PromiscReceiveCallback cb)
{
  m_promiscRx = cb;
}

bool
NeighborAwareDevice::SupportsSendFrom () const
{
  return false;
}
// }
std::vector<Ptr<NetDevice> >
NeighborAwareDevice::GetCommunicationNeighbors () const
{
  NS_ASSERT (m_mac != 0);
  return GetMac ()->GetCommunicationNeighbors ();
}

void
NeighborAwareDevice::ForwardUp (Ptr<Packet> packet, Address from, Address to, uint16_t protocolN)
{
  m_forwardUp (this, packet, protocolN, from);
}
}
}
